{*********************************************************}
{*                    RBNDB.PAS 2.11                     *}
{*       Copyright (c) Tamarack Associates 1996 - 2002.	 *}
{*                  All rights reserved.                 *}
{*       Modified with permission by Liodden Data AS     *}
{*********************************************************}

{$I RBDEFINE.INC}

unit rbNDB;

interface

uses

 Messages,
 SysUtils,
 Classes,
 Controls,
 DB,
 {$IFDEF HaveSysTools}
 StUtils,
 {$ENDIF}
 rbConst,
 rbBase,
 rbUtils,
 rbDS,
 rbUpdate,
 rbUpdDlg,
 rbSQL,
 NdbBase,
 NdbAsaDS,
 NdbApp,
 NdbUtil;

type

 TrbCustomWordsNDBLink = class(TrbWordsSQLLink)
 private
  function GetAsaDataset : TAsaDataset;
  procedure SetAsaDataset(Value : TAsaDataset);
 protected
  function GetDataSetInfo(Index : Integer) : String; override;
  procedure AddLikeIndex; override;
  procedure AddReverseIndex; override;
  procedure AddWordIndex; override;
  procedure CommitTransactions; override;
  procedure Properties(Stream : TStream ; Mode : TBlobStreamMode);
  procedure SetDataSetInfo(Index : Integer; const Value : String); override;
  procedure SetLike(const Value : String); override;
  procedure SetSegment(Value : Integer); override;
  procedure SetWord(const Value : String); override;
  procedure SetWordCount(Value : LongInt); override;
  procedure StartTransactions; override;
 public
  function GotoWord(const Word : String; Segment : Integer; ReadOnly : Boolean) : Boolean; override;
  procedure CreateDataSet; override;
  procedure Insert; override;
  procedure Post; override;
  procedure PrepareScan(S : String; AnyChar,OneChar : Char); override;
  procedure ReadProperties(Stream : TStream); override;
  procedure SaveData(var Data); override;
  procedure SaveProperties(Stream : TStream); override;
  property AsaDataSet : TAsaDataset read GetAsaDataset write SetAsaDataset;
 end;

 TrbMakeWordsNDBLink = class(TrbCustomWordsNDBLink)
 public
  constructor Create(AOwner : TComponent); override;
 published
  property AfterPost;
  property AsaDataSet;
  property CharFieldSize;
  property LikeFieldSize;
  property Params;
  property ReverseField;
  property TableName;
  property Version;                                                      
  property WordFieldSize;
 end;

 TrbWordsNDBLink = class(TrbCustomWordsNDBLink)
 public
  property CharFieldSize;
  property LikeFieldSize;
  property Params;
  property WordFieldSize;
 published
  property AfterPost;
  property AsaDataSet;
  property TableName;
  property Version;
 end;

 TrbCustomTextNDBLink = class(TrbTextSQLLink)
 private
  function GetAsaDataset : TAsaDataset;
  procedure SetAsaDataset(Value : TAsaDataset);
 protected
  function GetDataSetInfo(Index : Integer) : String; override;
  procedure SetDataSetInfo(Index : Integer; const Value : String); override;
  procedure SetIndexFieldName(Value : String); override;
  property AsaDataSet : TAsaDataSet read GetAsaDataset write SetAsaDataset;
 public
  function Eof : Boolean; override;
  function GotoLocation(ALocation : LongInt) : Boolean; override;
  function GotoNearestLocation(ALocation : LongInt) : Boolean; override;
  procedure GetFieldNames(List : TStrings); override;
  procedure GetIndexNames(List : TStrings; All : Boolean); override;
  procedure Highest; override;
  procedure Lowest; override;
 end;

 TrbTextNDBLink = class(TrbCustomTextNDBLink)
 public
  property Field;
 published
  property AfterProcess;
  property AsaDataSet;
  property BlockSize;
  property OnMaxIndex;
  property OnMinIndex;
  property OnProcessField;
  property SelectAll;
  property SubFieldNames;
  property TableName;
  property Version;
 end;

 TrbMakeTextNDBLink = class(TrbCustomTextNDBLink)
 public
  property Field;
 published
  property AfterProcess;
  property AsaDataSet;
  property BlockSize;
  property FieldNames;
  property IndexFieldName;
  property OnMaxIndex;
  property OnMinIndex;
  property OnProcessField;
  property SelectAll;
  property TableName;
  property Version;
 end;

 TrbAsaDataSet = class(TAsaDataset)
 private
  FAfterUpdate    : TNotifyEvent;
  FUpdater        : TrbBaseUpdate;
  FUpdatesEnabled : Boolean;
  {$IFDEF WIN32}
  procedure CMNotifyDataSet(var Message: TMessage); message CM_NOTIFYDATASET;
  {$ENDIF}
  procedure SetUpdater(Value : TrbBaseUpdate);
 protected
  procedure DoAfterDelete; override;
  procedure DoAfterPost; override;
  procedure DoAfterUpdate; virtual;
  procedure DoBeforeDelete; override;
  procedure DoBeforeEdit; override;
  procedure DoBeforeInsert; override;
  procedure Notification(AComponent: TComponent; Operation: TOperation); override;
 public
  constructor Create(AOwner : TComponent); override;
 published
  property AfterUpdate : TNotifyEvent read FAfterUpdate write FAfterUpdate;
  property Updater : TrbBaseUpdate read FUpdater write SetUpdater;
  property UpdatesEnabled : Boolean read FUpdatesEnabled write FUpdatesEnabled default True;
 end;

implementation

function TrbCustomWordsNDBLink.GetAsaDataset : TAsaDataset;
begin
 Result := TAsaDataset(DataSet)
end;

procedure TrbCustomWordsNDBLink.SetAsaDataset(Value : TAsaDataset);
begin
 SetDataSet(Value)
end;

function TrbCustomWordsNDBLink.GetDataSetInfo(Index : Integer) : String;
begin
 case Index of
  0 : if DataSet <> nil then
       Result := NdbExtractFilePrefix(ExtractFileName(AsaDataSet.Session.LoginDatabase))
      else
       Result := '';
  1 : Result := FTableName
 end
end;

procedure TrbCustomWordsNDBLink.AddLikeIndex;
begin
 if LikeFieldSize > 0 then
  with AsaDataSet do
   begin
    qAddLikeIndex(SQL);
    ExecSQL
   end
end;

procedure TrbCustomWordsNDBLink.AddReverseIndex;
begin
 if ReverseFieldSize > 0 then
  with AsaDataSet do
   begin
    qAddReverseIndex(SQL);
    ExecSQL
   end
end;

procedure TrbCustomWordsNDBLink.AddWordIndex;
begin
 with AsaDataSet do
  begin
   qAddWordIndex(SQL);
   ExecSQL
  end
end;

procedure TrbCustomWordsNDBLink.CommitTransactions;
begin
 if DataSet <> nil then
  begin
   AsaDataset.Session.Commit;
   inherited CommitTransactions
  end
end;

procedure TrbCustomWordsNDBLink.Properties(Stream : TStream; Mode : TBlobStreamMode);
begin
 Close;
 if DataSet <> nil then
  with AsaDataSet do
   try
    SQL.Clear;
    FLastQuery := wqProperties;
    SQL.Add('SELECT * FROM ' + TableName + ' WHERE ' + fnWord  + ' = :' + fnWord);
    ReadOnly := False;
    ParamByName(fnWord).AsString := PropertiesWord;
    Self.Open;
    if Mode = bmWrite then
     if Bof = Eof then
      begin
       Self.Close;
       ReadOnly := True;
       qProperties(SQL);
       Prepare;
       ParamByName(fnWord).AsString := PropertiesWord;
       ParamByName(fnSize).AsInteger := 0;
       ParamByName(fnBlob).LoadFromStream(Stream,ntBlob);
       if Segmented then
        ParamByName(fnSegment).AsInteger := 0;
       ExecSQL;
       CheckTransactions
      end
     else
      begin
       Edit;
       inherited SaveProperties(Stream);
       Post;
       CheckTransactions
      end
    else
     if Bof <> Eof then
      inherited ReadProperties(Stream)
     else
      RaiseRubiconError(rbePropertiesNotFound)
   finally
    Self.Close;
    SQL.Clear;
    ReadOnly := True
   end
end;

procedure TrbCustomWordsNDBLink.SetDataSetInfo(Index : Integer; const Value : String);
begin
 case Index of
  0 : if DataSet <> nil then
       AsaDataSet.Session.LoginDatabase := Value;
  1 : FTableName := Value
 end
end;

procedure TrbCustomWordsNDBLink.SetLike(const Value : String);
begin
 if LikeFieldSize > 0 then
  if Assigned(FFldLike) then
   FFldLike.AsString := Value
  else
   with AsaDataset do
    ParamByName(fnLike).AsString := Value
end;

procedure TrbCustomWordsNDBLink.SetSegment(Value : Integer);
begin
 if Segmented then
  if Assigned(FFldSegment) then
   FFldSegment.AsInteger := Value
  else
   with AsaDataset do
    ParamByName(fnSegment).AsInteger := Value
end;

procedure TrbCustomWordsNDBLink.SetWord(const Value : String);
begin
 if Assigned(FFldWord) then
  inherited SetWord(Value)
 else
  with AsaDataset do
   begin
    ParamByName(fnWord).AsString := Value;
    if ReverseFieldSize > 0 then
     ParamByName(fnReverse).AsString := ReverseStr(Value)
   end
end;

procedure TrbCustomWordsNDBLink.SetWordCount(Value : LongInt);
begin
 if Assigned(FFldWordCount) then
  FFldWordCount.AsInteger := Value
 else
  with AsaDataset do
   ParamByName(fnCount).AsInteger := Value
end;

procedure TrbCustomWordsNDBLink.StartTransactions;
begin
 inherited StartTransactions
end;

function TrbCustomWordsNDBLink.GotoWord(const Word : String; Segment : Integer;
                                        ReadOnly : Boolean) : Boolean;
begin
 Result := False;
 if DataSet <> nil then
   begin
    Close;
    AsaDataset.ReadOnly := ReadOnly;
    if ((not Segmented or (Segment < 0)) and (FLastQuery <> wqGotoWord1)) or
       ((Segmented and (Segment >= 0)) and (FLastQuery <> wqGotoWord2)) then
     with AsaDataSet do
      begin
       qGotoWord(SQL,Segment);
       Prepare
      end;
    with AsaDataset do
     begin
      Params[0].AsString := Word;
      if Segmented and
         (Segment >= 0) then
       Params[1].AsInteger := Segment
     end;
    Open;
    with DataSet do
     Result := Bof <> Eof
   end
end;

procedure TrbCustomWordsNDBLink.CreateDataSet;
begin
 Params.Values['Blob'] := 'long binary';
 with AsaDataset do
  begin
   qDropTable(SQL);
   try
    ExecSQL
   except
   end;
   qCreateTable(SQL);
   ExecSQL
  end
end;

procedure TrbCustomWordsNDBLink.Insert;
begin
 if Assigned(FFldWordCount) then                                         
  inherited Insert
 else                                                                    
  with AsaDataset do
   begin
    if FLastQuery <> wqInsert then
     begin
      qInsert(SQL);
      Prepare
     end;
    Inc(FInsertCounter)
   end;
end;

procedure TrbCustomWordsNDBLink.Post;
begin
 with AsaDataset do
  if FLastQuery = wqInsert then
   ExecSQL
  else
   begin
    Post;
    Self.Close;
   end;
 CheckTransactions;
 if Assigned(FAfterPost) then
  FAfterPost(Self)
end;

procedure TrbCustomWordsNDBLink.PrepareScan(S : String; AnyChar,OneChar : Char);
begin
 if DataSet <> nil then
  begin
   Close;
   qPrepareScan(AsaDataset.SQL,S,AnyChar,OneChar);
   Open;
   if IndexSize = 0 then
    Next
  end
end;

procedure TrbCustomWordsNDBLink.ReadProperties(Stream : TStream);
begin
 Properties(Stream,bmRead);
 if Assigned(FFldLike) then
  try
   {* Check if rbLike is indexed *}
  except
   FLikeFields := ''
  end
end;

procedure TrbCustomWordsNDBLink.SaveData(var Data);
var       BlobData     : PChar;
          Buffer       : AnsiString;
          BytesWritten : LongInt;
          Bytes        : LongInt;
begin
 if Assigned(FFldWordCount) then
  inherited SaveData(Data)
 else
  begin
   BlobData := @Data;
   Bytes := Abs(TCacheData(Data).Size);
   Dec(Bytes,TagSize);
   Inc(BlobData,TagSize);
   BytesWritten := 0;
   WordCount := Abs(TCacheData(Data).Count);
   AsaDataset.ParamByName(fnSize).AsInteger := Bytes;
   if FCharFieldSize > 0 then
    begin
     SetLength(Buffer,MinLong(FCharFieldSize,((Bytes * 4) div 3 + 3) and not 3));
     UUEncode(BlobData^,Buffer[1],Length(Buffer) * 3 div 4);
     AsaDataset.ParamByName(fnChar).AsString := Buffer;
     BytesWritten := MinLong(FCharFieldBytes,Bytes);
     Inc(BlobData,BytesWritten)
    end;
   with AsaDataset.ParamByName(fnBlob) do
    if Bytes > BytesWritten then
      SetBlobData(BlobData,Bytes - BytesWritten)
    else
     Clear;
   Inc(FBlobBytesWritten,Bytes)
  end
end;

procedure TrbCustomWordsNDBLink.SaveProperties(Stream : TStream);
begin
 if BooleanParam('SaveProperties',True) then
  Properties(Stream,bmWrite)
end;

constructor TrbMakeWordsNDBLink.Create(AOwner : TComponent);
begin
 inherited Create(AOwner);
 FCapabilities := FCapabilities + [cpBlob, cpChar];
 FIsMake := True
end;

function TrbCustomTextNDBLink.GetAsaDataset : TAsaDataset;
begin
 Result := TAsaDataset(DataSet)
end;

procedure TrbCustomTextNDBLink.SetAsaDataset(Value : TAsaDataset);
begin
 SetDataSet(Value)
end;

function TrbCustomTextNDBLink.GetDataSetInfo(Index : Integer) : String;
begin
 case Index of
  0 : if DataSet <> nil then
       Result :=  NdbExtractFilePrefix(ExtractFileName(AsaDataSet.Session.LoginDatabase))
      else
       Result := '';
  1 : Result := FTableName
 end
end;

procedure TrbCustomTextNDBLink.SetDataSetInfo(Index : Integer; const Value : String);
begin
 case Index of
  0 : if DataSet <> nil then
       AsaDataSet.Session.LoginDatabase := Value;
  1 : FTableName := Value
 end
end;

procedure TrbCustomTextNDBLink.SetIndexFieldName(Value : String);
begin
 inherited SetIndexFieldName(Value)
end;

function TrbCustomTextNDBLink.Eof : Boolean;
var      BlockStart : LongInt;
begin
 with TAsaDataset(SafeDataSet) do
  begin
   Result := Eof;
   BlockStart := Location;
   while Result and
         (BlockStart < MaxIndex) do
    begin
     Self.Close;
     PrepareQuery(SQL,tqBlock);
     if not Prepared then
      Prepare;
     Params[0].AsInteger := BlockStart;
     Params[1].AsInteger := BlockStart + FBlockSize;
     Self.Open;
     Result := Eof;
     Inc(BlockStart,FBlockSize)
    end
  end
end;

function TrbCustomTextNDBLink.GotoLocation(ALocation : LongInt) : Boolean;
begin
 Close;
 with TAsaDataset(SafeDataSet) do
  begin
   PrepareQuery(SQL,tqGoto);
   if not Prepared then
    Prepare;
   Params[0].AsInteger := ALocation;
   Self.Open;
   Result := Bof <> Eof
  end
end;

function TrbCustomTextNDBLink.GotoNearestLocation(ALocation : LongInt) : Boolean;
begin
 Close;
 with TAsaDataset(SafeDataSet) do
  begin
   PrepareQuery(SQL,tqBlock);
   if not Prepared then
    Prepare;
   Params[0].AsInteger := ALocation - 1;
   Params[1].AsInteger := ALocation + FBlockSize;
   Self.Open;
   Result := Bof <> Eof
  end
end;

procedure TrbCustomTextNDBLink.GetFieldNames(List : TStrings);
begin
 List.Clear;
 if DataSet <> nil then
  NdbGetFieldNames(AsaDataSet.Session.LoginDatabase, FTableName, List, TAsaDataset);
end;

procedure TrbCustomTextNDBLink.GetIndexNames(List : TStrings; All : Boolean);
begin
 List.Clear;
 if DataSet <> nil then
  with AsaDataSet do
   begin
    SQL.Clear;
    SQL.Add('select sx.index_name, sc.column_name, si."order"');
    SQL.Add('from sys.sysuserperms up, sys.systable st, sys.sysindex sx, sys.sysixcol si, sys.syscolumn sc');
    SQL.Add('where st.creator=up.user_id and st.table_id=si.table_id and sx.index_id=si.index_id and');
    SQL.Add('sx.table_id=si.table_id and si.table_id=sc.table_id and si.column_id=sc.column_id and st.table_name=:table_name');
    SQL.Add('order by si.index_id, si.sequence');
    ParamByName('table_name').AsString := FTableName;
    Open;
    try
     First;
     while not Eof do
      begin
       List.Add(FieldByName('column_name').AsString);
       Next
      end
    finally
     Close;
    end
   end
end;

procedure TrbCustomTextNDBLink.Highest;
var       MaxLocation : LongInt;
begin
 Close;
 with TAsaDataset(SafeDataSet) do
  begin
   if not Assigned(FOnMaxIndex) and
      (FMaxIndex = Undefined) then
    begin
     PrepareQuery(SQL,tqMax);
     Open;
     MaxLocation := Fields[0].AsInteger;
     Close;
    end
   else
    MaxLocation := MaxIndex;
   GotoLocation(MaxLocation)
  end
end;

procedure TrbCustomTextNDBLink.Lowest;
var       MinLocation : LongInt;
begin
 Close;
 with TAsaDataset(SafeDataSet) do
  begin
   if not Assigned(FOnMinIndex) and
      (FMinIndex = Undefined) then
    begin
     PrepareQuery(SQL,tqMin);
     Open;
     MinLocation := Fields[0].AsInteger;
     Close;
    end
   else
    MinLocation := MinIndex;
   if Assigned(FOnMinIndex) then
    if GotoNearestLocation(MinLocation) then
    else
     RaiseRubiconError(rbeBlockSizeTooSmall)
   else
    GotoLocation(MinLocation)
  end
end;

{ TrbAsaDataSet }

constructor TrbAsaDataSet.Create(AOwner : TComponent);
begin
 inherited Create(AOwner);
 FUpdatesEnabled := True
end;

{$IFDEF WIN32}
procedure TrbAsaDataSet.CMNotifyDataSet(var Message: TMessage);
begin
 if Assigned(FAfterUpdate) then
  Message.Result := 0
 else
  with TrbUpdateDialog(Message.LParam) do
   begin
    AfterUpdate := UpdateEvent;
    Updater := FUpdater;
    Message.Result := 1;
   end
end;
{$ENDIF}

procedure TrbAsaDataSet.SetUpdater(Value : TrbBaseUpdate);
begin
 FUpdater := Value;
end;

procedure TrbAsaDataSet.DoAfterDelete;
begin
 inherited DoAfterDelete;
 if Assigned(FUpdater) and
    FUpdatesEnabled then
  begin
   FUpdater.AfterDelete;
   DoAfterUpdate
  end
end;

procedure TrbAsaDataSet.DoAfterPost;
begin
 inherited DoAfterPost;
 if Assigned(FUpdater) and
    FUpdatesEnabled then
  begin
   FUpdater.AfterPost;
   DoAfterUpdate
  end;
end;

procedure TrbAsaDataSet.DoAfterUpdate;
begin
 if Assigned(FAfterUpdate) then
  FAfterUpdate(Self)
end;

procedure TrbAsaDataSet.DoBeforeDelete;
begin
 if Assigned(FUpdater) and
    FUpdatesEnabled then
  FUpdater.BeforeDelete;
 inherited DoBeforeDelete
end;

procedure TrbAsaDataSet.DoBeforeEdit;
begin
 if Assigned(FUpdater) and
    FUpdatesEnabled then
  FUpdater.BeforeEdit;
 inherited DoBeforeEdit
end;

procedure TrbAsaDataSet.DoBeforeInsert;
begin
 if Assigned(FUpdater) and
    FUpdatesEnabled then
  FUpdater.BeforeInsert;
 inherited DoBeforeInsert
end;

procedure TrbAsaDataSet.Notification(AComponent : TComponent ;
	   	        	    Operation : TOperation);
begin
 inherited Notification(AComponent, Operation);
 if Operation = opRemove then
  if AComponent = FUpdater then
   Updater := nil
end;

end.
